FileAttributes = [      # file attributes which can be used in queries
            "creator", "created_timestamp", "name", "namespace", "size"
]  

DatasetAttributes = [
    "creator", "created_timestamp", "namespace", "name",
    "description", "frozen", "monotonic"
]
